import struct
import os
import string


import DilloStuff
reload(DilloStuff)




dilloDir = r"C:\Program Files\Armadillo Run\Levels"

levelList = [
    "Empty.cost-rope-1.lvl",
    "Empty.cost-rope-2.lvl",
    "Empty.cost-rope-3.lvl",
    "Empty.cost-rope-4.lvl",
    "Empty.cost-rope-5.lvl",
    "Empty.cost-cloth-1.lvl",
    "Empty.cost-cloth-2.lvl",
    "Empty.cost-cloth-3.lvl",
    "Empty.cost-cloth-4.lvl",
    "Empty.cost-cloth-5.lvl",
    "Empty.cost-elastic-1.lvl",
    "Empty.cost-elastic-2.lvl",
    "Empty.cost-elastic-3.lvl",
    "Empty.cost-elastic-4.lvl",
    "Empty.cost-elastic-5.lvl",
]
levelList = os.listdir(dilloDir)
     



def checkHeader(level,sln):
    errors = []
    warnings = []

    #Check Flags
    if level.features&DilloStuff.FEATURE_FLAG_LEVEL==0:
        errors.append("Level flag not set on original level")
    if  sln.features&DilloStuff.FEATURE_FLAG_LEVEL==0:
        errors.append("Level flag not set on solution")
    if sln.features&DilloStuff.FEATURE_FLAG_SOLUTION==0:
        errors.append("Solution flag not set (probably modified using the editor, rather than the game)")

    #check background
    if level.background!=sln.background:
        warnings.append("Solution background (%d) does not match level background (%d)"%(sln.background,level.background))
        
    # check bounds
    if (level.leftBound != sln.leftBound) or (level.topBound != sln.topBound) or (level.rightBound != sln.rightBound )or (level.bottomBound != sln.bottomBound):
        errors.append("Solution bounds (%f,%f,%f,%f) do not match level bounds (%f,%f,%f,%f)"%(sln.leftBound,sln.topBound,sln.rightBound,sln.bottomBound,levelleftBound,level.topBound,level.rightBound,level.bottomBound))

    return (errors,warnings)

def checkBudget(level,sln):
    errors = []
    warnings = []

    spentBudget = 0
    for link in sln.links:
        if link.isUser():
            spentBudget+=link.cost()

    if spentBudget+sln.budget!=level.budget:
        errors.append("Budget mismatch: spent $%d + surplus $%d != $%d "%(spentBudget,sln.budget,level.budget))
    return (errors,warnings)

def checkDillos(level,sln):
    errors = []
    warnings = []

    #Check Dillos and Portals
    if len(level.dillos)!=len(sln.dillos):
        errors.append("Different number of Dillos")
    else:
        for i in range(0,len(level.dillos)):
            #Check for Dillo movement
            if (level.dillos[i].dilloX != sln.dillos[i].dilloX) or (level.dillos[i].dilloY != sln.dillos[i].dilloY):
                errors.append("'Dillo %d has been moved (%f,%f)->(%f,%f)"%(i,level.dillos[i].dilloX,level.dillos[i].dilloY,sln.dillos[i].dilloX,sln.dillos[i].dilloY))

            #Check for change in Dillo size            
            if level.dillos[i].dilloSize != sln.dillos[i].dilloSize:
                errors.append("Solution 'Dillo size (%f) does not match level 'Dillo Size (%f)"%(sln.dillos[i].dilloSize,level.dillos[i].dilloSize))

            #Check for portal movement                    
            if (level.dillos[i].portalX != sln.dillos[i].portalX) or (level.dillos[i].portalY != sln.dillos[i].portalY):
                errors.append("Dillo %d has been moved (%f,%f)->(%f,%f)"%(i,level.dillos[i].portalX,level.dillos[i].portalY,sln.dillos[i].portalX,sln.dillos[i].portalY))

            #Check for portal size
            if level.dillos[i].portalSize != sln.dillos[i].portalSize:
                errors.append("Solution portal size (%f) does not match level 'portal Size (%f)"%(sln.portals[i].portalSize,level.portals[i].portalSize)  )

    
    return (errors,warnings)

def checkChecksums(level,sln):
    errors = []
    warnings = []

    if level.levelChecksum!=sln.levelChecksum:
        errors.append("Level checksums do not match 0x%08X, 0x%08X"%(level.levelChecksum,sln.levelChecksum))

    if level.slnChecksum!=(0,0):
        errors.append("Level has solution checksum (probably edited using the game and then renamed)")

    if sln.slnChecksum==(0,0):
        errors.append("Missing solution checksum (probably modified using the editor, rather than the game)")
    
    
    return (errors,warnings)


def checkPoints(level,sln):
    def checkBounds(point,threshhold=0.1):
        level = point.level
        if point.x<level.leftBound-threshhold:
            return False
        if point.x>level.rightBound+threshhold:
            return False
        if point.y<level.bottomBound-threshhold:
            return False
        if point.y>level.topBound+threshhold:
            return False
        
        return True
    errors = []
    warnings = []

    levelFixedPointCount = 0
    slnFixedPointCount = 0
    userPoints = []
    #Check the points in the original level for iregularities
    for point in level.points:
        if point.isEditor():
            levelFixedPointCount+=1
            if not checkBounds(point):
                warnings.append("Fixed point #%d is outside the level bounds"%(point.index,))
        else:
            userPoints.append(point.index)

                    
    #flag any user points that were in the original
    if len(userPoints)>0:
        warnings.append("Level contains user point[s] [%s]"%string.join(map(lambda i:"#%d"%i,userPoints),","))
        
    for point in sln.points:
        if point.isEditor():
            slnFixedPointCount+=1

    #Check that the Solution points match the original level
    if levelFixedPointCount!=slnFixedPointCount:
        errors.append("Number of fixed points in level and solution do not match!")
    else:

        for i in range(0,levelFixedPointCount):
            levelPoint = level.points[i]
            slnPoint = sln.points[i]
            if (levelPoint.index!=slnPoint.index) or (levelPoint.type!=slnPoint.type) or (levelPoint.x!=slnPoint.x) or (levelPoint.y!=slnPoint.y):
                errors.append("Fixed point does not match corresponding point in original level (#%d %s (%f,%f)) != (#%d %s (%f,%f))"%(levelPoint.index,DilloStuff.PointNames[levelPoint.type],levelPoint.x,levelPoint.y,slnPoint.index,DilloStuff.PointNames[slnPoint.type],slnPoint.x,slnPoint.y))
        
    #Check that everything the user has added is inside the level
    for point in sln.points:
        if point.isUser():
            if not checkBounds(point):
                errors.append("User created point #%d is outside the level bounds"%(point.index,))
            
    
    return (errors,warnings)

def checkLinks(level,sln):
    errors = []
    warnings = []

    levelFixedLinkCount = 0
    slnFixedLinkCount = 0
    userLinks = []
    for link in level.links:
        if link .isEditor():
            levelFixedLinkCount+=1
        else:
            userLinks.append(link.index)

    if len(userLinks)>0:
        warnings.append("Level contains user link(s) [%s]"%string.join(map(lambda i:"#%d"%i,userLinks),","))
        
    for link in sln.links:
        if link.isEditor():
            slnFixedLinkCount+=1

    if levelFixedLinkCount!=slnFixedLinkCount:
        errors.append("Number of fixed Links in level and solution do not match!")
    else:

        for i in range(0,levelFixedLinkCount):
            levelLink = level.links[i]
            slnLink = sln.links[i]
            if (levelLink.index!=slnLink.index) or (levelLink.flags!=slnLink.flags) or (levelLink.type!=slnLink.type) or (levelLink.fromIndex!=slnLink.fromIndex) or (levelLink.toIndex!=slnLink.toIndex):
                fromPoint = level.points[levelLink.fromIndex]
                toPoint = level.points[levelLink.toIndex]
                levelLinkStr = "link(#%d,%s,%s,#%d(%f,%f)->#%d(%f,%f))"%(levelLink.index,DilloStuff.LinkFlagUserNames[levelLink.flags&DilloStuff.LINK_FLAGS_USER],DilloStuff.LinkNames[levelLink.type],fromPoint.index,fromPoint.x,fromPoint.y,toPoint.index,toPoint.x,toPoint.y)

                fromPoint = sln.points[slnLink.fromIndex]
                toPoint = sln.points[slnLink.toIndex]
                slnLinkStr = "link(#%d,%s,%s,#%d(%f,%f)->#%d(%f,%f))"%(slnLink.index,DilloStuff.LinkFlagUserNames[slnLink.flags&DilloStuff.LINK_FLAGS_USER],DilloStuff.LinkNames[slnLink.type],fromPoint.index,fromPoint.x,fromPoint.y,toPoint.index,toPoint.x,toPoint.y)

                errors.append("Fixed link does not match corresponding link in original level %s != %s)"%(levelLinkStr,slnLinkStr))


    #max length for each link type,  Note: these have some wiggle room to allow for floating point errors.
    maxLengthPerType = [
        None,
        None,
        2.002,
        2.002,
        None,
        2.002,
        3.002
    ]
        

    for link in level.links:
        linkLength = link.length()
        maxLength = maxLengthPerType[link.type]
        if maxLength!=None:
            if linkLength>maxLength:
                if link.isUser():
                    errors.append("User Link #%d is longer than the max length (%f > %f)"%(link.index,linkLength,maxLength))
                else:
                    warnings.append("Level Link #%d is longer than the max length (%f > %f)"%(link.index,linkLength,maxLength))

        
    
    return (errors,warnings)

def checkGroups(level,sl):
    errors = []
    warnings = []

    #Actually I can't think of a way to cheat using groups...

    return (errors,warnings)

def tern(a,b,c):
    if a:
        return b
    else:
        return c
        
totalErrors = 0
totalWarnings = 0
for filename in levelList:
    if os.path.splitext(filename)[1]==".lvl":
        sln = DilloStuff.DilloLevel()
        sln.load(os.path.join(dilloDir,filename))
        
        if sln.solutionName!="":

            #print filename
            
            errors = []
            warnings = []
            fullLevelName = os.path.join(dilloDir,sln.levelName+".lvl")
            if os.path.exists(fullLevelName):
                level = DilloStuff.DilloLevel()
                level.load(fullLevelName)

                checkFuncs = [
                    checkHeader,
                    checkBudget,
                    checkDillos,
                    checkChecksums,
                    checkPoints,
                    checkLinks,
                    checkGroups,
                ]

                for checkFunc in checkFuncs:
                    (funcErrors,funcWarnings)=checkFunc(level,sln)
                    errors +=funcErrors
                    warnings +=funcWarnings
                    
            else:
                errors.append("Original level file\"%s\" not found"%(sln.levelName+".lvl"))

            if len(errors)>0 or len(warnings)>0:
                print "%s (%d %s %d %s)"%(filename,len(errors),tern(len(errors)>1,"errors","error"),len(warnings),tern(len(warnings)>1,"warnings","warnings"))
                for warning in warnings:
                    print "\tWARNING: %s"%(warning)
                for error in errors:
                    print "\tERROR: %s"%(error)
            totalErrors += len(errors)                    
            totalWarnings += len(warnings)                    
print "TOTAL: (%d %s %d %s)"%(totalErrors,tern(totalErrors>1,"errors","error"),totalWarnings,tern(totalWarnings>1,"warnings","warnings"))
            
            
